# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: task.proto
# plugin: python-betterproto
from dataclasses import dataclass
from datetime import datetime

import betterproto

from .google import protobuf


@dataclass
class ProtoTaskMetadata(betterproto.Message):
    not_before: datetime = betterproto.message_field(1)
    expires: datetime = betterproto.message_field(2)
    max_retries: int = betterproto.uint32_field(3)
    attempt: int = betterproto.uint32_field(4)
    scheduled: datetime = betterproto.message_field(5)
    received: datetime = betterproto.message_field(6)
    queue_name: str = betterproto.string_field(7)
    extra: protobuf.Any = betterproto.message_field(10)


@dataclass
class ProtoTask(betterproto.Message):
    uuid: str = betterproto.string_field(1)
    meta: "ProtoTaskMetadata" = betterproto.message_field(2)
    task_type: str = betterproto.string_field(3)
    correlation_id: str = betterproto.string_field(4)
    payload: protobuf.Any = betterproto.message_field(10)
